%CMDILATION   Apply the dilation to an image (segments to segments)
%   segments = CMDILATION(segments,se_size)

% (C) Copyright 2002-2003               Pattern Recognition Group
%     All rights reserved               Faculty of Applied Physics
%                                       Delft University of Technology
%                                       Lorentzweg 1
%                                       2628 CJ Delft
%                                       The Netherlands
%
% Cris Luengo, August 2002.

function out = cmdilation(in,sesize)

% Test
if ~isstruct(in) | ~isfield(in,'coefficients') | ~isfield(in,'offsets')
   error('Expected a segment image.')
end
N = size(in.coefficients,2);
if N~=length(in.offsets)-1
   error('Segment image is not correct.')
end
if nargin < 2
   sesize = 5;
else
   if ~isnumeric(sesize) | prod(size(sesize))~=1 | sesize<1
      error('SESIZE must be a positive scalar (larger than 1).')
   end
end

r = sesize/2;
N = size(in.coefficients,2);
limits = in.offsets([1,end])+[-r,r];

% Create maxp signal with horizontal segments at each maximum
[mx,my] = cmlistmaxima(in);
if ~isempty(mx)
   % List of maxima in descending order
   [tmp,I] = sort(my);
   I = I(end:-1:1);
   my = my(I);
   mx = mx(I);

   % Polynome segments for maxima
   ii = 1;
   maxp = [0;0;0;my(ii);mx(ii)-r;mx(ii)+r];
   for ii=2:length(my)
      interval = mx(ii)+[-r,r];
      if interval(1) < limits(1)
         interval(1) = limits(1);
      end
      if interval(2) > limits(2)
         interval(2) = limits(2);
      end
      % Don't overlap with previous segments.
      % Note: this segment can never be split in two by an overlapping prevoius segment, since
      % they all start out with the same size.
      % We do this looping over the segments found so far. This way we don't make any
      % mistakes (it's an iterative process). Note that later segments are cut short by
      % earlier ones, so they might, in fact, try to cut ours into bits. By doing this
      % iteratively, this cannot happen. The code commented out below does not work as
      % advertised because of this.
      for jj=1:size(maxp,2)
         if interval(1) < maxp(5,jj)
            interval(2) = min(interval(2),maxp(5,jj));
         elseif interval(2) > maxp(6,jj)
            interval(1) = max(interval(1),maxp(6,jj));
         else
            % overlap!
            interval(2) = interval(1);
         end
      end
      %I = interval(1) < maxp(5,:);
      %if any(I)
      %   interval(2) = min(interval(2),min(maxp(5,I)));
      %end
      %I = ~I;
      %if any(I)
      %   interval(1) = max(interval(1),max(maxp(6,I)));
      %end
      if interval(1) < interval(2)
         maxp = [maxp,[0;0;0;my(ii);interval(1);interval(2)]];
      end
   end

   [tmp,I] = sort(maxp(5,:));
   maxp = maxp(:,I);

   % Insert -inf segments
   if maxp(5,1) > limits(1)
      maxp = [[0;0;0;-inf;limits(1);maxp(5,1)],maxp];
   end
   ii = 2;
   while ii <= size(maxp,2)
      if maxp(5,ii) > maxp(6,ii-1)
         maxp = [maxp(:,1:ii-1),[0;0;0;-inf;maxp(6,ii-1);maxp(5,ii)],maxp(:,ii:end)];
         ii = ii+1;
      end
      ii = ii+1;
   end
   if maxp(6,end) < limits(2)
      maxp = [maxp,[0;0;0;-inf;maxp(6,end);limits(2)]];
   end
   maxp = struct('coefficients',maxp(1:4,:),'offsets',[maxp(5,:),maxp(6,end)]);

else

   % There are no maxima. Weird, but possible.
   maxp = struct('coefficients',[0;0;0;-inf],'offsets',limits(:));

end

% Create shifted signals in1 and in2
leftmostvalue = in.coefficients(4,1);
rightmostvalue = cmevaluatepoly(in.coefficients(:,end),in.offsets(end)-in.offsets(end-1));
in1 = in;
in1.coefficients = [ [0;0;0;leftmostvalue] , in1.coefficients ];
in1.offsets = [ limits(1) , in1.offsets+r ];
in2 = in;
in2.coefficients = [ in2.coefficients , [0;0;0;rightmostvalue] ];
in2.offsets = [ in2.offsets-r, limits(2) ];

% Take maximum of { maxp, in(x-r), in(x+r) }
out = maximumsegments(maxp,in1); % maxp is the simplest function: result is simple
out = maximumsegments(out,in2);  % Here the result is complicated a bit more.
                                 % Doing maximumsegments(in1,in2) first would create a very
                                 % complex function



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Returns the maximum of the two input functions
function out = maximumsegments(in1,in2)
N1 = size(in1.coefficients,2);
N2 = size(in2.coefficients,2);
in1I = 1; % index into in1
in2I = 1; % index into in2
region = [0,0];
region(1) = in1.offsets(1); % leftmost edge. Must be equal for both functions.
if region(1) ~= in2.offsets(1)
   error('Assertion failed (1)')
end
out = struct('coefficients',[],'offsets',region(1));
while 1
   region(2) = min(in1.offsets(in1I+1),in2.offsets(in2I+1));
   % Determine if there are intersection points.
   a1 = in1.coefficients(:,in1I);
   offset1 = region(1) - in1.offsets(in1I);
   if offset1~=0
      a1 = shiftpoly(a1,-offset1);
   end
   a2 = in2.coefficients(:,in2I);
   offset2 = region(1) - in2.offsets(in2I);
   if offset2~=0
      a2 = shiftpoly(a2,-offset2);
   end
   if any(~isfinite(a1))
      % special case 1
      out.coefficients = [ out.coefficients , a2 ];
      out.offsets = [ out.offsets , region(2) ];
   elseif any(~isfinite(a2))
      % special case 2
      out.coefficients = [ out.coefficients , a1 ];
      out.offsets = [ out.offsets , region(2) ];
   elseif a1==a2
      % special case 3
      out.coefficients = [ out.coefficients , a1 ];
      out.offsets = [ out.offsets , region(2) ];
   else
      % find intersection points, don't bother about sections that are too small...
      x = findzeros(a1-a2,[0,region(2)-region(1)]) + region(1);
      if ~isempty(x)
         if x(1)-region(1) < 1e-6 % very small!!!!
            x(1) = [];
         end
      end
      if ~isempty(x)
         if region(2)-x(end) < 1e-6 % very small!!!!
            x(end) = [];
         end
      end
      if ~isempty(x)
         I = find(diff(x) < 1e-6); % very small!!!!
         if any(I)
            x(I) = [];
         end
      end
      x = [region(1),x,region(2)];
      % append largest of each sub-segment : we're creating lots of segments here!
      for ii=1:length(x)-1
         % avaluating the two polynomes halfway the sub-segment we discover which one is larger.
         % this is easier than to compare the derivatives and the starting value and stuff...
         if cmevaluatepoly(a1,(x(ii)+x(ii+1))/2-region(1)) > cmevaluatepoly(a2,(x(ii)+x(ii+1))/2-region(1))
            v1 = cmevaluatepoly(a1,x(ii)-region(1));
            v2 = cmevaluatepoly(a1,x(ii+1)-region(1));
            offset = x(ii)-region(1);
            %if offset~=0
            %   tmpa = shiftpoly(a1,-offset);
            %else
            %   tmpa = a1;
            %end
            out.coefficients = [ out.coefficients , shiftpoly(a1,-offset) ];
         else
            v1 = cmevaluatepoly(a2,x(ii)-region(1));
            v2 = cmevaluatepoly(a2,x(ii+1)-region(1));
            offset = x(ii)-region(1);
            %if offset~=0
            %   tmpa = shiftpoly(a2,-offset);
            %else
            %   tmpa = a2;
            %end
            out.coefficients = [ out.coefficients , shiftpoly(a2,-offset) ];
         end
         out.offsets = [ out.offsets , x(ii+1) ];
      end
   end
   % next segments
   region(1) = region(2);
   if in1.offsets(in1I+1) <= region(1)
      in1I = in1I+1;
   end
   if in2.offsets(in2I+1) <= region(1)
      in2I = in2I+1;
   end
   if (in1I > N1) & (in2I > N2)
      return
   end
   if (in1I > N1) | (in2I > N2)
      [in1I,N1]
      [in2I,N2]
      [in1.offsets(end-1:end);in2.offsets(end-1:end);[0,region(1)]]
      error('Assertion failed (2)')
   end
end



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Solves a cubic equation. Clips the result to the region.
% x can be empty, or have 3 values...
function x = findzeros(coef,region)
if coef(1) == 0
   if coef(2) == 0
      if coef(3) == 0
         x = [];
      else
         % linear equation
         x = -coef(4)/coef(3);
      end
   else
      % cuadratic equation
      a = coef(2); b = coef(3); c = coef(4);
      D = b^2 - 4*a*c;
      if D>0
         sqrtD = sqrt(D);
         x = (-b+[sqrtD,-sqrtD]) / a;
      elseif D==0
         x = -b/a;
      else
         x = [];
      end
   end
else
   % cubic equation
   coef = coef/coef(1);
   a = coef(2); b = coef(3); c = coef(4);
   Q = (a^2-3*b)/9;
   R = (2*a^3-9*a*b+27*c)/54;
   Q3 = Q^3;
   R2 = R^2;
   if R2<Q3
      Theta = acos(R/sqrt(Q3));
      x = -2*sqrt(Q)*cos((Theta+2*pi*[-1,0,1])/3) - a/3;
   else
      A = -sign(R)*(abs(R)+sqrt(R2-Q3))^(1/3);
      if A==0
         x = -a/3;
      else
         x = A + Q/A - a/3;
      end
   end
end
if ~isempty(x)
   x = sort(x);
   x(x<=region(1)) = [];
end
if ~isempty(x)
   x(x>=region(2)) = [];
end



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Shifts a polynome by d
% returns the coefficients of  (x-r).^[3,2,1,0] * a
function out = shiftpoly(in,d)
A = [ 1,0,0,0 ; -3*d,1,0,0 ; 3*d^2,-2*d,1,0 ; -d^3,d^2,-d,1 ];
out = A*in;
